from planning_sat.pddl_adapter import PlanningProblem
from pddlpy.pddl import Operator


initial_state = [("adjacent", "loc1", "loc2"),
                 ("adjacent", "loc2", "loc1"),
                 ("in", "conta", "loc1"),
                 ("in", "contb", "loc2"),
                 ("atl", "robr", "loc1"),
                 ("atl", "robq", "loc2"),
                 ("unloaded", "robr"),
                 ("unloaded", "robq")]

goal_state = [("in", "contb", "loc1"),
              ("in", "conta", "loc2")]

fluents = [('atl', 'robr', 'loc1'),
           ('atl', 'robr', 'loc2'),
           ('atl', 'robq', 'loc1'),
           ('atl', 'robq', 'loc2'),
           ('loaded', 'robr', 'conta'),
           ('loaded', 'robr', 'contb'),
           ('loaded', 'robq', 'conta'),
           ('loaded', 'robq', 'contb'),
           ('unloaded', 'robr'),
           ('unloaded', 'robq'),
           ('in', 'conta', 'loc1'),
           ('in', 'conta', 'loc2'),
           ('in', 'contb', 'loc1'),
           ('in', 'contb', 'loc2')]


class TestPlanningProblem:

    def test_constructor(self):
        pp = PlanningProblem("domain/dock-worker-robot-domain.pddl",
                             "domain/dock-worker-robot-problem.pddl")

        assert type(pp.initial_state) is set
        assert type(pp.goal_state) is set
        assert type(pp.actions) is list
        assert type(pp.fluents) is list

    def test_initial_state(self):
        pp = PlanningProblem("domain/dock-worker-robot-domain.pddl",
                             "domain/dock-worker-robot-problem.pddl")
        assert pp.initial_state == set(initial_state)

    def test_goal_state(self):
        pp = PlanningProblem("domain/dock-worker-robot-domain.pddl",
                             "domain/dock-worker-robot-problem.pddl")
        assert pp.goal_state == set(goal_state)

    def test_actions(self):
        pp = PlanningProblem("domain/dock-worker-robot-domain.pddl",
                             "domain/dock-worker-robot-problem.pddl")
        for act in pp.actions:
            assert type(act) is Operator

    def test_fluents(self):
        pp = PlanningProblem("domain/dock-worker-robot-domain.pddl",
                             "domain/dock-worker-robot-problem.pddl")
        assert set(pp.fluents) == set(fluents)



"""Encoder

Description:
    This module runs DPLL (Davis–Putnam–Logemann–Loveland) Algorithm
    to solve satisfiability problem

License:
    Copyright 2021 Debby Nirwan

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from planning_sat.encoder import PlanningProblemEncoder, Operator, Clause
from typing import List, Set
import copy
import pddlpy
import time
import argparse
import os
import sys


class DPLL(object):

    def __call__(self, formulas: List[Clause], model=None):

        if len(formulas) == 0:
            return True, model

        if any([clause.empty for clause in formulas]):
            return False, None

        if not model:
            model = set()

        next_clause, unit = self._select_literal(formulas)

        if not unit or ('not' not in next_clause):
            positive_formulas = copy.deepcopy(formulas)
            positive_clause = self._pos_clause(next_clause)
            new_model, new_formulas = self._unit_propagation(positive_formulas,
                                                             model,
                                                             positive_clause)

            sat, new_model = self.__call__(new_formulas, new_model)
            if sat:
                return sat, new_model

        if not unit or ('not' in next_clause):
            negative_formulas = copy.deepcopy(formulas)
            negative_clause = self._neg_clause(next_clause)
            new_model, new_formulas = self._unit_propagation(negative_formulas,
                                                             model,
                                                             negative_clause)

            sat, new_model = self.__call__(new_formulas, new_model)
            if sat:
                return sat, new_model

        return False, None

    def _select_literal(self, formulas):
        if formulas == list():
            return None, None

        unit_clause = self._get_unit_clause(formulas)
        if unit_clause:
            return unit_clause, True

        return formulas[0].clause[0], False

    @staticmethod
    def _get_unit_clause(formulas):
        for clause in formulas:
            if clause.is_single:
                return clause.clause[0]
        return None

    @staticmethod
    def _unit_propagation(formulas: List[Clause], model: Set[Clause],
                          unit_clause):

        new_formulas = copy.deepcopy(formulas)
        model = model.union(unit_clause)

        if not new_formulas:
            return model, new_formulas

        if 'not' in unit_clause.clause[0]:
            reverse_clause = list(unit_clause.clause[0])
            reverse_clause.remove('not')
            not_unit_clause = Clause(tuple(reverse_clause))
        else:
            reverse_clause = ('not',) + unit_clause.clause[0]
            not_unit_clause = Clause(reverse_clause)

        for clause in list(new_formulas):
            for cl in clause.clause:
                if isinstance(cl, Operator):
                    continue
                if unit_clause.clause[0] == cl:
                    new_formulas.remove(clause)
                    break
                if not_unit_clause.clause[0] == cl:
                    new_formulas.remove(clause)
                    new_clause = Clause()
                    literals = []
                    for literal in clause:
                        if not isinstance(literal, Operator) and \
                                literal != not_unit_clause.clause[0]:
                            literals.append(literal)
                    if literals:
                        for literal in literals:
                            new_clause.add(literal, Operator.OR)
                    new_formulas.append(new_clause)
                    break

        return model, new_formulas

    @staticmethod
    def _pos_clause(atom: tuple):
        if 'not' in atom:
            pos_atom = list(atom)
            pos_atom.remove('not')
            pos_atom = tuple(pos_atom)
        else:
            pos_atom = atom

        return Clause(pos_atom)

    @staticmethod
    def _neg_clause(atom: tuple):
        if 'not' in atom:
            neg_atom = atom
        else:
            neg_atom = ('not',) + atom

        return Clause(neg_atom)

    @staticmethod
    def _pos_and_neg_atoms(atom: tuple):
        if 'not' in atom:
            neg_atom = atom
            pos_atom = list(atom)
            pos_atom.remove('not')
            pos_atom = tuple(pos_atom)
        else:
            pos_atom = atom
            neg_atom = ('not',) + atom

        return pos_atom, neg_atom


def setup_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Solving SAT Formulas using DPLL Algorithm",
        allow_abbrev=False
    )

    parser.add_argument(
        "-d", "--domain",
        required=True,
        type=str,
        help="path to pddl domain file"
    )

    parser.add_argument(
        "-p", "--problem",
        required=True,
        type=str,
        help="path to pddl problem file"
    )

    parser.add_argument(
        "-l", "--length",
        required=True,
        type=int,
        default=1,
        help="specify the length of steps in formulas encoding"
    )

    parser.add_argument(
        "-f", "--print",
        action='store_true',
        help="print the result"
    )

    return parser


if __name__ == "__main__":

    args = setup_parser().parse_args()
    if not os.path.isfile(args.domain) or not os.path.isfile(args.problem):
        sys.exit(1)

    domain_file = args.domain
    problem_file = args.problem
    length = args.length
    print_debug = args.print

    pp_encoder = PlanningProblemEncoder(domain_file, problem_file, length)

    dpll = DPLL()

    if print_debug:
        print(f"DPLL algorithm running with "
              f"{len(pp_encoder.propositional_formulas)} formulas")
    start_time = time.perf_counter()
    result_dp, final_model = dpll(pp_encoder.propositional_formulas)
    end_time = time.perf_counter()
    if print_debug:
        print(f"DPLL algorithm ran for {end_time-start_time:0.4f} seconds")
        if result_dp:
            print("Plan:")
            operator_list = []
            for item in final_model:
                if 'not' not in item and isinstance(item[0], pddlpy.Operator):
                    operator_list.append(item)
            operator_list.sort(key=lambda tup: tup[-1])
            for op in operator_list:
                print(op)
        else:
            print(f"Failed to plan at length {length}")



from planning_sat.dpll import DPLL
from planning_sat.encoder import Clause, Operator


model = {('not', 'B'), ('not', 'A'), ('D',)}


class TestDPLL:

    def test_dpll(self):
        forms = []
        clause_1 = Clause(('D',))
        forms.append(clause_1)
        clause_2 = Clause(('not', 'D'))
        clause_2.add(('A',), Operator.OR)
        clause_2.add(('not', 'B'), Operator.OR)
        forms.append(clause_2)
        clause_3 = Clause(('not', 'D'))
        clause_3.add(('not', 'A'), Operator.OR)
        clause_3.add(('not', 'B'), Operator.OR)
        forms.append(clause_3)
        clause_4 = Clause(('not', 'D'))
        clause_4.add(('not', 'A'), Operator.OR)
        clause_4.add(('B',), Operator.OR)
        forms.append(clause_4)
        clause_5 = Clause(('D',))
        clause_5.add(('A',), Operator.OR)
        forms.append(clause_5)

        dpll = DPLL()
        result_dp, final_model = dpll(forms)

        assert result_dp
        assert final_model == model



"""Encoder

Description:
    This module encodes Planning Problem to Propositional Formulas in CNF
    (Conjunctive Normal Form)

License:
    Copyright 2021 Debby Nirwan

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from .pddl_adapter import PlanningProblem
from enum import Enum
from itertools import combinations


class Operator(Enum):
    AND = 0,
    OR = 1,
    IMPLIES = 2


class Clause(object):

    def __init__(self, fluent=None):
        if fluent:
            self._clause = [fluent]
            self._single = True
        else:
            self._clause = []
            self._single = False

    def __repr__(self):
        return f"Clause object. {self._clause}"

    def __len__(self):
        return len(self._clause)

    def __getitem__(self, item):
        return self._clause[item]

    def __contains__(self, item):
        return True if item in self._clause else False

    def __eq__(self, other):
        if self._single == other.is_single and self._clause == other.clause:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def add(self, fluent, operator: Operator):
        if len(self._clause) == 0:
            self._single = True
        else:
            self._single = False
            self._clause.append(operator)
        self._clause.append(fluent)

        return self

    @property
    def clause(self):
        return self._clause

    @property
    def is_single(self):
        return self._single

    @property
    def empty(self):
        return self._clause == []


class PlanningProblemEncoder(object):

    def __init__(self, dom_file: str, problem_file: str, length=1):
        self._problem = PlanningProblem(dom_file, problem_file)
        self._length = length
        self._propositional_formulas = self._encode()

    def _encode(self):
        actions = self._problem.actions
        fluents = self._problem.fluents

        # 1. encode initial state
        init_state = list(self._problem.initial_state)
        init_state_clauses = []
        for fluent in list(fluents):
            if fluent not in init_state:
                fluent = ('not',) + fluent
            fluent = fluent + ('0',)
            init_state_clauses.append(Clause(fluent))

        # 2. encode goal state
        goal_state = list(self._problem.goal_state)
        goal_state_clauses = []
        for goal in goal_state:
            goal_state_clauses.append(Clause(goal + (str(self._length),)))

        enc_actions_clauses = []
        explanatory_frame_axioms = []
        complete_exclusion_axiom = []

        for step in range(self._length):
            # 3. encode actions
            for act in actions:
                if act.effect_pos.issubset(act.precondition_pos):
                    continue
                action_tuple = ('not', act, str(step))
                # preconditions
                for p in act.precondition_pos:
                    if 'adjacent' in p:
                        continue
                    action_clause = Clause(action_tuple)
                    p = p + (str(step),)
                    action_clause.add(p, Operator.OR)
                    enc_actions_clauses.append(action_clause)
                # positive effects
                for e in act.effect_pos:
                    e = e + (str(step + 1),)
                    action_clause = Clause(action_tuple)
                    action_clause.add(e, Operator.OR)
                    enc_actions_clauses.append(action_clause)
                # negative effects
                for e in act.effect_neg:
                    e = ('not',) + e + (str(step + 1),)
                    action_clause = Clause(action_tuple)
                    action_clause.add(e, Operator.OR)
                    enc_actions_clauses.append(action_clause)

            # 4. explanatory frame axioms
            for fluent in fluents:
                act_with_pos_effect = []
                act_with_neg_effect = []
                for act in actions:
                    if act.effect_pos.issubset(act.precondition_pos):
                        continue
                    if fluent in act.effect_pos:
                        act_with_pos_effect.append(act)
                    elif fluent in act.effect_neg:
                        act_with_neg_effect.append(act)
                if act_with_pos_effect:
                    a_pos = fluent + (str(step),)
                    b_pos = ('not',) + fluent + (str(step + 1),)
                    clause_pos = Clause(a_pos)
                    clause_pos.add(b_pos, Operator.OR)
                    for act in act_with_pos_effect:
                        c_pos = (act, str(step))
                        clause_pos.add(c_pos, Operator.OR)
                    explanatory_frame_axioms.append(clause_pos)
                if act_with_neg_effect:
                    a_neg = ('not',) + fluent + (str(step),)
                    b_neg = fluent + (str(step + 1),)
                    clause_neg = Clause(a_neg)
                    clause_neg.add(b_neg, Operator.OR)
                    for act in act_with_neg_effect:
                        c_neg = (act, str(step))
                        clause_neg.add(c_neg, Operator.OR)
                    explanatory_frame_axioms.append(clause_neg)

            # 5. complete exclusion axiom
            for action_pair in combinations(actions, 2):
                if action_pair[0].effect_pos.issubset(
                        action_pair[0].precondition_pos):
                    continue
                if action_pair[1].effect_pos.issubset(
                        action_pair[1].precondition_pos):
                    continue
                action0_tuple = ('not', action_pair[0], str(step))
                action1_tuple = ('not', action_pair[1], str(step))
                action_pair_clause = Clause(action0_tuple)
                action_pair_clause.add(action1_tuple, Operator.OR)
                complete_exclusion_axiom.append(action_pair_clause)

        proposition_formulas = init_state_clauses + goal_state_clauses + \
            enc_actions_clauses + explanatory_frame_axioms + \
            complete_exclusion_axiom

        return proposition_formulas

    @property
    def propositional_formulas(self):
        return self._propositional_formulas
